var _ = require("underscore");
var crc32 = require('buffer-crc32');
var Client = require('./client');
var EventEmitter = require('events');
var util = require('util');

var CONTAINNER_COMM_PING = "Container.Common.Ping";
var CRC32_CONTAINNER_COMM_PING = crc32.unsigned(CONTAINNER_COMM_PING);
var DEFAULT_OPT = {
    host: "127.0.0.1",
    port: 80,
    name: '',
    num: 5,
    heartbeat: 30 * 1000
};

function ClientPool(opt) {
    opt = _.extend({}, DEFAULT_OPT, opt);

    this._host = opt.host;
    this._port = opt.port;
    this._name = opt.name;
    this._number = opt.num;
    this._heartbeat = opt.heartbeat;
    this._count = 0;
    this._timer = null;
    this._poolObj = [];
    this._queue = [];
    this._isInit = true;

    this.on('ready', _.bind(this._onReady, this));
    this._createItems(this._number);
}

util.inherits(ClientPool, EventEmitter);

ClientPool.prototype._createItems = function (num) {
    var _me = this;
    this._count += num;

    var host = this._host;
    var port = this._port;
    var name = this._name;

    for (var i = 0; i < num; i++) {
        (function () {
            var client = new Client(host, port, name);
            client.on('connect', function () {
                client.register();
            });

            client.on('error', function (error) {
                _me.emit('error', error);
                _me._count -= 1;
                _me._createItems(1);
                client.destroy();
            });

            client.on("register", function () {
                _me._poolObj.push(client);

                client.heartbeat = +new Date();
                client.isiDle = true;
                client.decoder.on("message", _.bind(_me._onMessage, client));

                if (_me._count >= _me._number) {
                    _me._isInit && _me._checkIdle();
                    _me._isInit = false;
                    _me.emit('ready');
                }
            });
        }());
    }
};

ClientPool.prototype._onReady = function () {
    if (!this._queue.length) {
        return;
    }

    var minLen = Math.min(this._queue.length, this._poolObj.length);
    for (var i = 0; i < minLen; i++) {
        this.acquire(this._queue.shift());
    }
};

ClientPool.prototype._onMessage = function (msg) {
    var serviceName = msg.service;
    if (serviceName == CRC32_CONTAINNER_COMM_PING) {
        this.heartbeat = +new Date();
    }
};

ClientPool.prototype._checkIdle = function () {
    var _me = this;
    var now = +new Date();
    clearTimeout(this._timer);
    this._timer = setTimeout(function () {
        var createNum = 0;
        for (var i = _me._poolObj.length - 1; i >= 0; i--) {
            var item = _me._poolObj[i];
            if (item.isiDle && now - item.heartbeat >= _me._heartbeat) {
                _me._poolObj.splice(i, 1);
                item.destroy();
                createNum += 1;
            }
        }

        if (createNum) {
            _me._count -= createNum;
            _me._createItems(createNum);
        }

        _me._checkIdle();
    }, this._heartbeat);
};

ClientPool.prototype.acquire = function (callback) {
    if (!this._poolObj.length) {
        this._queue.push(callback);
    } else {
        var item = this._poolObj.shift();
        item.isiDle = false;
        callback(item);
    }
};

ClientPool.prototype.release = function (client) {
    if (client) {
        client.isiDle = true;
        this._poolObj.push(client);
        this._onReady();
    }
};

ClientPool.prototype.destroy = function () {
    for (var i = 0, len = this._poolObj.length; i < len; i++) {
        this._poolObj[i].destroy();
    }

    clearTimeout(this._timer);
    this._count = 0;
    this._timer = null;
    this._poolObj = [];
    this._queue = [];

    this.removeAllListeners('ready');
};

module.exports = ClientPool;