var util = require('util');
var Socket = require('net').Socket;
var EventEmitter = require('events');
var getmac = require('getmac');
var crc32 = require('buffer-crc32');
var thenjs = require('thenjs');
var _ = require('underscore');
var BigNumber = require('bignumber.js');

var Encoder = require('./encoder');
var Decoder = require('./decoder');

var NO_DELAY = true;
var SOCKET_TIMEOUT = 0;
var CONTAINNER_COMM_PING = "Container.Common.Ping";
var CONTAINNER_COMM_REGISTER = "Container.Service.Register";
var CRC32_CONTAINNER_COMM_PING = crc32.unsigned(CONTAINNER_COMM_PING);
var CRC32_CONTAINNER_COMM_REGISTER = crc32.unsigned(CONTAINNER_COMM_REGISTER);
var FLAG = require('./flag');

function Service(host, port, name) {
    EventEmitter.call(this);
    this.name = name;
    this.rand = Math.ceil(Math.random() * 65536);
    this.closed = false;
    this.uniqid = 0;
    this.askId = 0;
    this.encoder = new Encoder();
    this.decoder = new Decoder();
    this.client = new Socket({
        readable: true,
        writable: true
    });

    this.client.setNoDelay(NO_DELAY);
    this.client.setTimeout(SOCKET_TIMEOUT);
    this.client.on('connect', _.bind(this._onConnectHandler, this));
    this.client.on('close', _.bind(this._onCloseHandler, this));
    this.client.on('data', _.bind(this._onDataHandler, this));
    this.client.on('error', _.bind(this._onErrorHandler, this));
    this.client.connect({
        host: host,
        port: port
    });

    this.decoder.on('message', _.bind(this._onMessage, this));
    this.decoder.on('error', _.bind(this._onErrorHandler, this));

}

util.inherits(Service, EventEmitter);

Service.prototype.register = function () {
    var _me = this;
    var body = {};

    thenjs(function (cont) {
        getmac.getMac(function (error, macAddress) {
            cont(error, macAddress);
        });
    }).then(function (cont, macAddress) {
        var macInts = macAddress.split(/:|\-/);
        if (macInts.length < 6) {
            throw new Error('cant get mac address');
        }

        var pid = process.pid + _me.rand;
        body.strtag = new BigNumber('0x' + macInts.join('') + (pid % 65536).toString(16)).toString(10);
        body.name = _me.name || 'service_' + body.strtag;
        body.accepts = _.map(_me.accepts, function (val, key) {
            return {
                service: key,
                broadcast: val.broadcast,
                force: val.force
            }
        });
        cont();
    }).then(function () {
        var msg = _me.encoder.encode(body, _me._genMsg(CONTAINNER_COMM_REGISTER));
        _me.client.write(msg);
    }).fail(function (cont, error) {
        // TODO log error
        console.log(error);
    });
};

Service.prototype.rpcRequest = function (body, msg) {
    msg = this.encoder.encode(body, msg);
    this.client.write(msg);
};

Service.prototype.eventRequest = function (serviceName, body) {
    var msg = this._genMsg(serviceName, {
        flag: FLAG.EVENT
    });
    msg = this.encoder.encode(body, msg);
    this.client.write(msg);
};

Service.prototype.close = function () {
    if (!this.closed) {
        this.closed = true;
        this.client.destroy(); //ensure no i/o activies
    }
};

Service.prototype.destroy = function () {
    this.close();
    this.decoder.removeAllListeners('message');
    this.encoder = null;
    this.decoder = null;
    this.client = null;
};

Service.prototype._onConnectHandler = function () {
    this.emit('connect');
};

Service.prototype._onCloseHandler = function () {
    this.emit('close');
};

Service.prototype._onDataHandler = function (buffer) {
    this.decoder.decode(buffer);
};

Service.prototype._onErrorHandler = function (error) {
    this.emit('error', error);
};

Service.prototype._onMessage = function (msg) {
    var serviceName = msg.service;
    var body = msg.body;
    var isCalled = false;

    if (serviceName == CRC32_CONTAINNER_COMM_PING) {
        this.eventRequest(CONTAINNER_COMM_PING, body);
        isCalled = true;
    } else if (serviceName == CRC32_CONTAINNER_COMM_REGISTER) {
        this.emit('register');
        isCalled = true;
    }

    return isCalled;
};

Service.prototype._genMsg = function (serviceName, opt) {
    return _.extend({
        service: crc32.unsigned(serviceName),
        time: 0,
        uniqid: 0,
        askId: this.askId++
    }, opt);
};

module.exports = Service;